/*
 * MP3.h
 *
 * public header for shared library
 *
 * This wizard-generated code is based on code adapted from the
 * SampleLib project distributed as part of the Palm OS SDK 4.0,
 *
 * Copyright (c) 1994-1999 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 */

#ifndef MP3_H_
#define MP3_H_

/* Palm OS common definitions */
#include <SystemMgr.h>

#include "config.h"
#include "decoder.h"

/* If we're actually compiling the library code, then we need to
 * eliminate the trap glue that would otherwise be generated from
 * this header file in order to prevent compiler errors in CW Pro 2. */
#ifdef BUILDING_MP3
	#define MP3_LIB_TRAP(trapNum)
#else
	#define MP3_LIB_TRAP(trapNum) SYS_TRAP(trapNum)
#endif

/*********************************************************************
 * Type and creator of Sample Library database
 *********************************************************************/

#define		MP3CreatorID	'hSm3'
#define		MP3TypeID		sysFileTLibrary

/*********************************************************************
 * Internal library name which can be passed to SysLibFind()
 *********************************************************************/

#define		MP3Name		"HS_MP3"

#define		MP3LibraryVer	1

/*********************************************************************
 * HS_MP3 result codes
 * (appErrorClass is reserved for 3rd party apps/libraries.
 * It is defined in SystemMgr.h)
 *********************************************************************/

/* invalid parameter */
#define MP3ErrParam		(appErrorClass | 1)		

/* library is not open */
#define MP3ErrNotOpen		(appErrorClass | 2)		

/* returned from MP3Close() if the library is still open */
#define MP3ErrStillOpen	(appErrorClass | 3)		

/* MAD signaled an error */
#define MP3ErrMAD		(appErrorClass | 4)

/*********************************************************************
 * API Prototypes
 *********************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

/* Standard library open, close, sleep and wake functions */

extern Err MP3Open(UInt16 refNum, UInt32 * clientContextP)
	MP3_LIB_TRAP(sysLibTrapOpen);
				
extern Err MP3Close(UInt16 refNum, UInt32 clientContext)
	MP3_LIB_TRAP(sysLibTrapClose);

extern Err MP3Sleep(UInt16 refNum)
	MP3_LIB_TRAP(sysLibTrapSleep);

extern Err MP3Wake(UInt16 refNum)
	MP3_LIB_TRAP(sysLibTrapWake);

/* Custom library API functions */

extern UInt16 MP3GetLibraryVer(UInt16 refNum)
	MP3_LIB_TRAP(sysLibTrapBase + 5);

extern Err MP3StreamBuffer(UInt16 refNum)
	MP3_LIB_TRAP(sysLibTrapBase + 6);

extern Err MP3DecoderInit(UInt16 refNum, struct mad_decoder *decoder,
		void *data,
		enum mad_flow (*input_func)(void *, struct mad_stream *),
		enum mad_flow (*header_func)(void *, struct mad_header const *),
		enum mad_flow (*filter_func)(void *, struct mad_stream const *,
				   struct mad_frame *),
		enum mad_flow (*output_func)(void *, struct mad_header const *,
				   struct mad_pcm *),
		enum mad_flow (*error_func)(void *, struct mad_stream *,
				  struct mad_frame *),
		enum mad_flow (*message_func)(void *, void *, unsigned int *))
	MP3_LIB_TRAP(sysLibTrapBase + 7);

extern Err MP3DecoderRun(UInt16 refNum, struct mad_decoder *decoder,
		enum mad_decoder_mode mode)
	MP3_LIB_TRAP(sysLibTrapBase + 8);

extern Err MP3DecoderFinish(UInt16 refNum, struct mad_decoder *decoder)
	MP3_LIB_TRAP(sysLibTrapBase + 9);

#ifdef __cplusplus
}
#endif

/*
 * FUNCTION: MP3_OpenLibrary
 *
 * DESCRIPTION:
 *
 * User-level call to open the library.  This inline function
 * handles the messy task of finding or loading the library
 * and calling its open function, including handling cleanup
 * if the library could not be opened.
 * 
 * PARAMETERS:
 *
 * refNumP
 *		Pointer to UInt16 variable that will hold the new
 *      library reference number for use in later calls
 *
 * clientContextP
 *		pointer to variable for returning client context.  The client context is
 *		used to maintain client-specific data for multiple client support.  The 
 *		value returned here will be used as a parameter for other library 
 *		functions which require a client context.  
 *
 * CALLED BY: System
 *
 * RETURNS:
 *		errNone
 *		memErrNotEnoughSpace
 *      sysErrLibNotFound
 *      sysErrNoFreeRAM
 *      sysErrNoFreeLibSlots
 *
 * SIDE EFFECTS:
 *		*clientContextP will be set to client context on success, or zero on
 *      error.
 */
 
__inline Err MP3_OpenLibrary(UInt16 *refNumP, UInt32 * clientContextP)
{
	Err error;
	Boolean loaded = false;
	
	/* first try to find the library */
	error = SysLibFind(MP3Name, refNumP);
	
	/* If not found, load the library instead */
	if (error == sysErrLibNotFound)
	{
		error = SysLibLoad(MP3TypeID, MP3CreatorID, refNumP);
		loaded = true;
	}
	
	if (error == errNone)
	{
		error = MP3Open(*refNumP, clientContextP);
		if (error != errNone)
		{
			if (loaded)
			{
				SysLibRemove(*refNumP);
			}

			*refNumP = sysInvalidRefNum;
		}
	}
	
	return error;
}

/*
 * FUNCTION: MP3_CloseLibrary
 *
 * DESCRIPTION:	
 *
 * User-level call to closes the shared library.  This handles removal
 * of the library from system if there are no users remaining.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number obtained from MP3_OpenLibrary().
 *
 * clientContext
 *		client context (as returned by the open call)
 *
 * CALLED BY: Whoever wants to close the library
 *
 * RETURNS:
 *		errNone
 *		sysErrParamErr
 */

__inline Err MP3_CloseLibrary(UInt16 refNum, UInt32 clientContext)
{
	Err error;
	
	if (refNum == sysInvalidRefNum)
	{
		return sysErrParamErr;
	}

	error = MP3Close(refNum, clientContext);

	if (error == errNone)
	{
		/* no users left, so unload library */
		SysLibRemove(refNum);
	} 
	else if (error == MP3ErrStillOpen)
	{
		/* don't unload library, but mask "still open" from caller  */
		error = errNone;
	}
	
	return error;
}

#endif /* MP3_H_ */
