/*
 * MTImpl.c
 *
 * implementation for shared library
 *
 * This wizard-generated code is based on code adapted from the
 * SampleLib project distributed as part of the Palm OS SDK 4.0.
 *
 * Copyright (c) 1994-1999 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 */

/* Our library public definitions (library API) */
#define BUILDING_MT
#include "MT.h"
#include "MTPrivate.h"
#include "layer3.h"

/*********************************************************************
 *
 * LIBRARY GLOBALS:
 *
 * IMPORTANT:
 * ==========
 * Libraries are *not* allowed to have global or static variables.
 * Instead, they allocate a memory chunk to hold their persistent data, 
 * and save a handle to it in the library's system library table entry.
 * Example functions below demostrate how the library "globals" chunk 
 * is set up, saved, and accessed.
 *
 * We use a movable memory chunk for our library globals to minimize
 * dynamic heap fragmentation.  Our library globals are locked only
 * when needed and unlocked as soon as possible.  This behavior is
 * critical for healthy system performance.
 *
 *********************************************************************/


/*********************************************************************
 * Internally used routines
 *********************************************************************/

static MTGlobalsType * PrvMakeGlobals(UInt16 refNum);
static void PrvFreeGlobals(UInt16 refNum);
static MTGlobalsType * PrvLockGlobals(UInt16 refNum);
static Boolean PrvIsLibOpen(UInt16 refNum);
static Err PrvCreateClientContext(MTGlobalsType * gP, UInt32 * clientContextP);
static Err PrvDestroyClientContext(MTGlobalsType * gP, UInt32 clientContext);
static MTClientContextType * PrvLockContext(UInt32 context);

/*********************************************************************
 * Internally used macros
 *********************************************************************/

/* Unlock globals */
#define PrvUnlockGlobals(gP)        MemPtrUnlock(gP)

/* Unlock the client context */
#define PrvUnlockContext(contextP)	MemPtrUnlock(contextP)
 
/*********************************************************************
 * Library API Routines
 *********************************************************************/
 
/*
 * FUNCTION: MTOpen
 *
 * DESCRIPTION:
 *
 * Opens the library, creates and initializes the globals.
 * This function must be called before any other library functions.
 *
 * If MTOpen fails, do not call any other library API functions.
 * If MTOpen succeeds, call MTClose when you are done using
 * the library to enable it to release critical system resources.
 *
 * LIBRARY DEVELOPER NOTES:
 *
 * The library's "open" and "close" functions should *not* take an excessive
 * amount of time to complete.  If the processing time for either of these
 * is lengthy, consider creating additional library API function(s) to handle
 * the time-consuming chores.
 *
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * clientContextP
 *		pointer to variable for returning client context.  The client context is
 *		used to maintain client-specific data for multiple client support.  The 
 *		value returned here will be used as a parameter for other library 
 *		functions which require a client context.  
 *
 * CALLED BY: anyone who wants to use this library
 *
 * RETURNS:
 *		errNone
 *		memErrNotEnoughSpace
 *
 * SIDE EFFECTS:
 *		*clientContextP will be set to client context on success, or zero on error.
 */

Err MTOpen(UInt16 refNum, UInt32 *clientContextP)
{
	MTGlobalsType *gP;
	Err err = errNone;
	Int16 originalOpenCount = 0;

	/* Error-check our parameters */
	ErrFatalDisplayIf(
		clientContextP == NULL, 
		"null context variable pointer");

	/* Initialize return variable */
	*clientContextP = 0;

	/* Get library globals */
	gP = PrvLockGlobals(refNum);

	/* Check if already open */
	if (!gP)
	{
		/* Allocate and initialize our library globals. */
		gP = PrvMakeGlobals(refNum);
		if ( !gP )
			err = memErrNotEnoughSpace;
	}

	/* If we have globals, create a client context, increment open
	 * count, and unlock our globals */
	if ( gP )
	{
		originalOpenCount = gP->openCount;

		err = PrvCreateClientContext(gP, clientContextP);
		if ( !err )
			gP->openCount++;

		PrvUnlockGlobals(gP);

		/* If there was an error creating a client context and there  */
		/* are no other clients, free our globals */
		if ( err && (originalOpenCount == 0) )
			PrvFreeGlobals(refNum);
	}

	return( err );
}

/*
 * FUNCTION: MTClose
 *
 * DESCRIPTION:	
 *
 * Closes the library, frees client context and globals.
 *
 * ***IMPORTANT***
 * May be called only if MTOpen succeeded.
 *
 * If other applications still have the library open, decrements
 * the reference count and returns MTErrStillOpen.
 *
 * LIBRARY DEVELOPER NOTES:
 *
 * The library's "open" and "close" functions should *not* take an excessive
 * amount of time to complete.  If the processing time for either of these
 * is lengthy, consider creating additional library API function(s) to handle
 * the time-consuming chores.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * clientContext
 *		client context
 *
 * CALLED BY: Whoever wants to close the library
 *
 * RETURNS:
 *		errNone
 *		MTErrStillOpen -- library is still open by others (no error)
 */

Err MTClose(UInt16 refNum, UInt32 clientContext)
{
	MTGlobalsType * gP;
	Int16 openCount;
	Int16 contextCount;
	Err err = errNone;

	gP = PrvLockGlobals(refNum);

	/* If not open, return */
	if (!gP)
	{
		/* MUST return zero here to get around a bug in system v1.x that
		 * would cause SysLibRemove to fail. */
		return errNone;
	}

	/* Destroy the client context (we ignore the return code in this implementation) */
	PrvDestroyClientContext(gP, clientContext);

	/* Decrement our library open count */
	gP->openCount--;

	/* Error check for open count underflow */
	ErrFatalDisplayIf(gP->openCount < 0, "HS_MP3_Tables open count underflow");

	/* Save the new open count and the context count */
	openCount = gP->openCount;
	contextCount = gP->contextCount;

	PrvUnlockGlobals(gP);

	/* If open count reached zero, free our library globals */
	if ( openCount <= 0 )
	{
		/* Error check to make sure that all client contexts were destroyed */
		ErrFatalDisplayIf(contextCount != 0, "not all client contexts were destroyed");

		/* Free our library globals */
		PrvFreeGlobals(refNum);
	}
	else
	{
		/* return this error code to inform the caller
		 * that others are still using this library */
		err = MTErrStillOpen;
	}

	return err;
}

/*
 * FUNCTION: MTSleep
 *
 * DESCRIPTION:
 *
 * Handles system sleep notification.
 *
 * ***IMPORTANT***
 * This notification function is called from a system interrupt.
 * It is only allowed to use system services which are interrupt-
 * safe.  Presently, this is limited to EvtEnqueueKey, SysDisableInts,
 * SysRestoreStatus.  Because it is called from an interrupt,
 * it must *not* take a long time to complete to preserve system
 * integrity.  The intention is to allow system-level libraries
 * to disable hardware components to conserve power while the system
 * is asleep.
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: System
 *
 * RETURNS:	errNone
 */

Err MTSleep(UInt16 refNum)
{
	#pragma unused(refNum)
	return errNone;
}

/*
 * FUNCTION: MTWake
 *
 * DESCRIPTION:
 *
 * Handles system wake notification.
 *
 * ***IMPORTANT***
 * This notification function is called from a system interrupt.
 * It is only allowed to use system services which are interrupt-
 * safe.  Presently, this is limited to EvtEnqueueKey, SysDisableInts,
 * SysRestoreStatus.  Because it is called from an interrupt,
 * it must *not* take a long time to complete to preserve system
 * integrity.  The intention is to allow system-level libraries
 * to enable hardware components which were disabled when the system
 * went to sleep.
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: System
 *
 * RETURNS:	errNone
 */

Err MTWake(UInt16 refNum)
{
	#pragma unused(refNum)
	return errNone;
}

/*
 * FUNCTION: MTGetTablesLE
 *
 * DESCRIPTION:
 *
 * *** DESCRIBE YOUR IMPLEMENTATION DETAILS ***
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

Err MTGetTablesLE(UInt16 refNum, MTTablesPtr tablesP)
{
	#pragma unused(refNum)
    return errNone;
}    

/*
 * FUNCTION: MTGetTablesBE
 *
 * DESCRIPTION:
 *
 * *** DESCRIBE YOUR IMPLEMENTATION DETAILS ***
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

#define alloc_copy(field) \
	if ((tablesP->field = MemPtrNew(sizeof(field))) == NULL) \
		return memErrNotEnoughSpace; \
	MemMove(tablesP->field, field, sizeof(field))

Err MTGetTablesBE(UInt16 refNum, MTTablesPtr tablesP)
{
	#pragma unused(refNum)
	
	alloc_copy(sflen_table);
	alloc_copy(nsfb_table);
	
	alloc_copy(sfb_48000_long);
	alloc_copy(sfb_44100_long);
	alloc_copy(sfb_32000_long);
	alloc_copy(sfb_48000_short);
	alloc_copy(sfb_44100_short);
	alloc_copy(sfb_32000_short);
	alloc_copy(sfb_48000_mixed);
	alloc_copy(sfb_44100_mixed);
	alloc_copy(sfb_32000_mixed);
	alloc_copy(sfb_24000_long);
	alloc_copy(sfb_22050_long);
	alloc_copy(sfb_24000_short);
	alloc_copy(sfb_22050_short);
	alloc_copy(sfb_16000_short);
	alloc_copy(sfb_24000_mixed);
	alloc_copy(sfb_22050_mixed);
	alloc_copy(sfb_16000_mixed);
	alloc_copy(sfb_8000_long);
	alloc_copy(sfb_8000_short);
	alloc_copy(sfb_8000_mixed);
	
	{
		struct sfbwidth_table_type (*tmp)[9];
		
		if ((tmp = MemPtrNew(sizeof(*tmp))) == NULL)
			return memErrNotEnoughSpace;
		
		(*tmp)[0].l = sfb_48000_long;
		(*tmp)[0].s = sfb_48000_short;
		(*tmp)[0].m = sfb_48000_mixed;
		(*tmp)[1].l = sfb_44100_long;
		(*tmp)[1].s = sfb_44100_short;
		(*tmp)[1].m = sfb_44100_mixed;
		(*tmp)[2].l = sfb_32000_long;
		(*tmp)[2].s = sfb_32000_short;
		(*tmp)[2].m = sfb_32000_mixed;
		(*tmp)[3].l = sfb_24000_long;
		(*tmp)[3].s = sfb_24000_short;
		(*tmp)[3].m = sfb_24000_mixed;
		(*tmp)[4].l = sfb_22050_long;
		(*tmp)[4].s = sfb_22050_short;
		(*tmp)[4].m = sfb_22050_mixed;
		(*tmp)[5].l = sfb_16000_long;
		(*tmp)[5].s = sfb_16000_short;
		(*tmp)[5].m = sfb_16000_mixed;
		(*tmp)[6].l = sfb_12000_long;
		(*tmp)[6].s = sfb_12000_short;
		(*tmp)[6].m = sfb_12000_mixed;
		(*tmp)[7].l = sfb_11025_long;
		(*tmp)[7].s = sfb_11025_short;
		(*tmp)[7].m = sfb_11025_mixed;
		(*tmp)[8].l = sfb_8000_long;
		(*tmp)[8].s = sfb_8000_short;
		(*tmp)[8].m = sfb_8000_mixed;
		
		tablesP->sfbwidth_table = tmp;
	}
	
	alloc_copy(pretab);
	alloc_copy(rq_table);
	alloc_copy(root_table);
	alloc_copy(cs);
	alloc_copy(ca);
	alloc_copy(imdct_s);
	alloc_copy(window_l);
	alloc_copy(window_s);
	alloc_copy(is_table);
	alloc_copy(is_lsf_table);
	
    return errNone;
}    

/*
 * FUNCTION: MTFreeTables
 *
 * DESCRIPTION:
 *
 * *** DESCRIBE YOUR IMPLEMENTATION DETAILS ***
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

Err MTFreeTables(UInt16 refNum, MTTablesPtr tablesP)
{
	#pragma unused(refNum)
	
	if (!tablesP)
		return MTErrParam;
	
	MemPtrFree(tablesP->sflen_table);
	MemPtrFree(tablesP->nsfb_table);
	
	MemPtrFree(tablesP->sfb_48000_long);
	MemPtrFree(tablesP->sfb_44100_long);
	MemPtrFree(tablesP->sfb_32000_long);
	MemPtrFree(tablesP->sfb_48000_short);
	MemPtrFree(tablesP->sfb_44100_short);
	MemPtrFree(tablesP->sfb_32000_short);
	MemPtrFree(tablesP->sfb_48000_mixed);
	MemPtrFree(tablesP->sfb_44100_mixed);
	MemPtrFree(tablesP->sfb_32000_mixed);
	MemPtrFree(tablesP->sfb_24000_long);
	MemPtrFree(tablesP->sfb_22050_long);
	MemPtrFree(tablesP->sfb_24000_short);
	MemPtrFree(tablesP->sfb_22050_short);
	MemPtrFree(tablesP->sfb_16000_short);
	MemPtrFree(tablesP->sfb_24000_mixed);
	MemPtrFree(tablesP->sfb_22050_mixed);
	MemPtrFree(tablesP->sfb_16000_mixed);
	MemPtrFree(tablesP->sfb_8000_long);
	MemPtrFree(tablesP->sfb_8000_short);
	MemPtrFree(tablesP->sfb_8000_mixed);
	
	MemPtrFree(tablesP->sfbwidth_table);
	MemPtrFree(tablesP->pretab);
	MemPtrFree(tablesP->rq_table);
	MemPtrFree(tablesP->root_table);
	MemPtrFree(tablesP->cs);
	MemPtrFree(tablesP->ca);
	MemPtrFree(tablesP->imdct_s);
	MemPtrFree(tablesP->window_l);
	MemPtrFree(tablesP->window_s);
	MemPtrFree(tablesP->is_table);
	MemPtrFree(tablesP->is_lsf_table);
	
    return errNone;
}    

/*********************************************************************
 * Private Functions
 *********************************************************************/

#pragma mark -

/*
 * FUNCTION: PrvMakeGlobals
 *
 * DESCRIPTION: Create our library globals.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *
 *		pointer to our *locked* library globals
 *		NULL if our globals	could not be created.
 */

static MTGlobalsType * PrvMakeGlobals(UInt16 refNum)
{
	MTGlobalsType * gP = NULL;
	MemHandle gH;
	SysLibTblEntryType * libEntryP;

	/* Get library globals */
	libEntryP = SysLibTblEntry(refNum);
	ErrFatalDisplayIf(libEntryP == NULL, "invalid HS_MP3_Tables refNum");

	/* Error check to make sure the globals don't already exist */
	ErrFatalDisplayIf(libEntryP->globalsP, "HS_MP3_Tables globals already exist");

	/* Allocate and initialize our library globals. */
	gH = MemHandleNew(sizeof(MTGlobalsType));
	if ( !gH )
		return( NULL );

	/* Save the handle of our library globals in the system library table  */
	/* entry so we can later retrieve it using SysLibTblEntry(). */
	libEntryP->globalsP = (void*)gH;

	/* Lock our globals (should not fail) */
	gP = PrvLockGlobals(refNum);
	ErrFatalDisplayIf(gP == NULL, "failed to lock HS_MP3_Tables globals");

	/* Set the owner of our globals memory chunk to "system" (zero), so it won't get
	 * freed automatically by Memory Manager when the first application to call
	 * MTOpen exits.  This is important if the library is going to stay open
	 * between apps. */
	MemPtrSetOwner(gP, 0);

	/* Initialize our library globals */
	MemSet(gP, sizeof(MTGlobalsType), 0);

	/* for convenience and debugging, save ref in globals structure */
	gP->thisLibRefNum = refNum;

	/* initial open count */
	gP->openCount = 0;

	/* return a pointer to our *locked* globals */
	return( gP );
}

/*
 * FUNCTION: PrvFreeGlobals
 *
 * DESCRIPTION: Free our library globals.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS: nothing
 */

static void PrvFreeGlobals(UInt16 refNum)
{
	MemHandle gH;
	SysLibTblEntryType * libEntryP;

	/* Get our library globals handle */
	libEntryP = SysLibTblEntry(refNum);
	ErrFatalDisplayIf(libEntryP == NULL, "invalid HS_MP3_Tables refNum");

	gH = (MemHandle)(libEntryP->globalsP);

	/* Free our library globals */
	if ( gH )
	{
		/* clear our globals reference */
		libEntryP->globalsP = NULL;

		/* free our globals */
		MemHandleFree(gH);
	}
}

/*
 * FUNCTION: PrvLockGlobals
 *
 * DESCRIPTION:	Lock our library globals
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		pointer to our library globals
 *		NULL if our globals	have not been created yet.
 */

static MTGlobalsType * PrvLockGlobals(UInt16 refNum)
{
	MTGlobalsType * gP = NULL;
	MemHandle gH;
	SysLibTblEntryType * libEntryP;

	libEntryP = SysLibTblEntry(refNum);
	if ( libEntryP )
		gH = (MemHandle)(libEntryP->globalsP);
	if ( gH )
		gP = (MTGlobalsType *)MemHandleLock(gH);

	return( gP );
}

/*
 * FUNCTION: PrvIsLibOpen
 *
 * DESCRIPTION:	Check if the library has been opened.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS: non-zero if the library has been opened
 */

static Boolean PrvIsLibOpen(UInt16 refNum)
{
	MTGlobalsType * gP;
	Boolean	isOpen = false;

	gP = PrvLockGlobals(refNum);

	if ( gP )
	{
		isOpen = true;
		PrvUnlockGlobals(gP);
	}

	return( isOpen );
}

/*
 * FUNCTION: PrvCreateClientContext
 *
 * DESCRIPTION:	
 *
 * Create a client context for storing client-specific data.
 * The client context allows the library to support multiple clients.
 *
 * PARAMETERS:
 *
 * gP
 *		pointer to our locked globals
 *
 * clientContextP
 *		pointer to variable for returning client context
 *
 * CALLED BY: Internal
 *
 * RETURNS:
 *		errNone
 *		memErrNotEnoughSpace
 *		MTErrNotOpen -- the library is not open
 *
 * SIDE EFFECTS:
 *
 *		*clientContextP will be set to client context on success, or zero on error.
 */

static Err PrvCreateClientContext(MTGlobalsType * gP, UInt32 * clientContextP)
{
	Err err = errNone;
	MemHandle contextH;
	MTClientContextType * contextP;

	/* Error-check our parameters */
	ErrFatalDisplayIf(gP == NULL, "null globals pointer");
	ErrFatalDisplayIf(clientContextP == NULL, "null context variable pointer");

	/* Initialize return variable */
	*clientContextP = 0;

	/* Allocate a new client context structure */
	contextH = MemHandleNew(sizeof(MTClientContextType));
	if ( !contextH )
	{
		err = memErrNotEnoughSpace;
	}
	else
	{
		/* save context chunk handle in return variable */
		*clientContextP = (UInt32)contextH;

		/* Initialize the context chunk */
		contextP = (MTClientContextType *)MemHandleLock(contextH);

		/* save address of open routine as signature to validate context */
		contextP->libSignature = (void *)&MTOpen;

		/* TODO: Insert code to initialize context members */

		PrvUnlockContext(contextP);

		/* increment context count (for debugging) */
		gP->contextCount++;
		ErrFatalDisplayIf(gP->contextCount == 0, "context count overflow");
	}

	return( err );
}

/*
 * FUNCTION: PrvDestroyClientContext
 *
 * DESCRIPTION:	Destroy a client context which was created by PrvCreateClientContext.
 *
 * PARAMETERS:
 *
 * gP
 *		pointer to our locked globals
 *
 * clientContextP
 *		client context
 *
 * CALLED BY: Internal
 *
 * RETURNS:
 *		errNone
 *		MTErrNotOpen -- the library is not open
 */

static Err PrvDestroyClientContext(MTGlobalsType * gP, UInt32 clientContext)
{
	MTClientContextType *	contextP;

	/* Error-check our parameters */
	ErrFatalDisplayIf(gP == NULL, "null globals pointer");

	/* Validate the client context by locking it */
	contextP = PrvLockContext(clientContext);

	if ( contextP )
	{
		/* freeing a locked chunk is permitted by the system */
		MemPtrFree(contextP);

		/* decrement context count (for debugging) */
		gP->contextCount--;
		ErrFatalDisplayIf(gP->contextCount < 0, "context count underflow");
	}

	return errNone;
}

/*
 * FUNCTION: PrvLockContext
 *
 * DESCRIPTION:	Validate and lock a client context.
 *
 * PARAMETERS: 
 *
 * context
 *		a client context to lock
 *
 * CALLED BY: internal
 *
 * RETURNS: pointer to the locked client context.
 */
 
static MTClientContextType * PrvLockContext(UInt32 context)
{
	MTClientContextType *	contextP = NULL;

	/* Error-check our parameters */
	ErrFatalDisplayIf(context == 0, "null client context");

	/* Lock the client context */
	contextP = (MTClientContextType *)MemHandleLock((MemHandle)context);
	ErrFatalDisplayIf(contextP == NULL, "failed to lock client context");

	/* Validate the client context */
	ErrFatalDisplayIf(contextP->libSignature != (void *)&MTOpen,
		"invalid client context");

	return( contextP );
}